from amber_mcap.dataset.conversion import decode_image_message, decode_message
from amber_mcap.dataset.rosbag2_dataset import Rosbag2Dataset, MessageMetaData
from amber_mcap.dataset.topic_config import ImageTopicConfig, TfTopicConfig
from amber_mcap.exception import TaskDescriptionError, RuntimeError
from amber_mcap.unit.time import Time, TimeUnit
from amber_mcap.util.geometry import build_tf_buffer
from amber_mcap.tf2_amber import (
    BufferCore,
    timeFromSec,
    displayTimePoint,
    durationFromSec,
    TransformStamped,
)
from dataclass_wizard import YAMLWizard
from dataclasses import dataclass, field
from mcap.reader import NonSeekingReader
from typing import Any, List, Optional, Tuple
import numpy as np
import datetime
import quaternion
import torch


@dataclass
class ReadImagesConfig(YAMLWizard):  # type: ignore
    image_topics: List[ImageTopicConfig] = field(default_factory=list)
    compressed: bool = True
    tf_topic: TfTopicConfig = TfTopicConfig()

    def get_image_topics(self) -> List[str]:
        topics: List[str] = []
        for topic in self.image_topics:
            topics.append(topic.topic_name)
        return topics

    def get_camera_info_topics(self) -> List[str]:
        topics: List[str] = []
        for topic in self.image_topics:
            if topic.camera_info_topic_name != None:
                topics.append(topic.camera_info_topic_name)
        return topics


class ImagesDataset(Rosbag2Dataset):  # type: ignore
    # images: List[torch.Tensor] = []
    num_images = 0
    config: ReadImagesConfig = ReadImagesConfig()
    tf_buffer: Optional[BufferCore] = None

    def __init__(
        self,
        rosbag_path: str,
        config: ReadImagesConfig,
        transform: Any = None,
        target_transform: Any = None,
    ) -> None:
        # self.images.clear()
        self.config = config
        print(self.config)
        super().__init__(
            rosbag_path,
            self.config.compressed,
            transform,
            target_transform,
        )
        self.tf_buffer, _, _ = build_tf_buffer(
            self.rosbag_files, self.config.tf_topic, self.config.compressed
        )
        self.count_images()

    def count_images(self) -> None:
        self.num_images = 0
        for rosbag_file in self.rosbag_files:
            reader = NonSeekingReader(rosbag_file)
            for schema, channel, message in reader.iter_messages():
                if channel.topic in self.config.get_image_topics():
                    self.num_images = self.num_images + 1
                    self.message_metadata.append(
                        MessageMetaData.from_dict(
                            {
                                "publish_time": datetime.datetime.fromtimestamp(
                                    Time(message.publish_time, TimeUnit.NANOSECOND).get(
                                        TimeUnit.SECOND
                                    ),
                                    tz=datetime.timezone.utc,
                                ),
                                "topic": channel.topic,
                                "rosbag_path": rosbag_file,
                            }
                        )
                    )

    def __len__(self) -> int:
        return self.num_images

    def __iter__(self) -> torch.Tensor:
        for rosbag_file in self.rosbag_files:
            reader = NonSeekingReader(rosbag_file)
            for schema, channel, message in reader.iter_messages():
                if channel.topic in self.config.get_image_topics():
                    yield decode_image_message(message, schema, self.config.compressed)

    def get_camera_info(self):
        if (
            len(self.config.get_image_topics()) != 1
            or len(self.config.get_camera_info_topics()) != 1
        ):
            raise TaskDescriptionError(
                "Multiple camera with camera_info does not supported."
            )
        for rosbag_file in self.rosbag_files:
            reader = NonSeekingReader(rosbag_file)
            for schema, channel, message in reader.iter_messages():
                if channel.topic in self.config.get_camera_info_topics():
                    return decode_message(message, schema, self.config.compressed)

    def get_camera_image_by_index(self, index: int):
        if index >= self.num_images or index < 0:
            raise RuntimeError("Index of the image was invalid. Index : " + str(index))
        current_index = 0
        for rosbag_file in self.rosbag_files:
            reader = NonSeekingReader(rosbag_file)
            for schema, channel, message in reader.iter_messages():
                if channel.topic in self.config.get_image_topics():
                    if current_index == index:
                        return decode_image_message(
                            message, schema, self.config.compressed
                        )
                    current_index = current_index + 1
        raise RuntimeError(
            "Failed to get camera image by index. Index "
            + str(index)
            + " was not found."
        )

    def transform_3d_point_to_image_coordinate(
        self,
        index: int,
        point_3d: Tuple[float, float, float],
        map_frame_id: str = "map",
    ) -> Tuple[float, float]:
        camera_info = self.get_camera_info()
        transform = self.tf_buffer.lookupTransform(
            map_frame_id,
            camera_info.header.frame_id,
            timeFromSec(self.get_metadata(index).publish_time.timestamp()),
        )
        rotation_quaternion = np.quaternion(
            transform.transform.rotation.w,
            transform.transform.rotation.x,
            transform.transform.rotation.y,
            transform.transform.rotation.z,
        ).conjugate()
        translation_vector = np.array(
            [
                transform.transform.translation.x,
                transform.transform.translation.y,
                transform.transform.translation.z,
            ],
            dtype=float,
        ).reshape(3)
        rotated_point = -quaternion.rotate_vectors(
            rotation_quaternion, np.array(point_3d, dtype=float).reshape(3)
        )
        rotated_translation = quaternion.rotate_vectors(
            rotation_quaternion, translation_vector
        )
        transformed_point = rotated_point + rotated_translation
        # # See also https://github.com/ros-perception/vision_opencv/blob/27de9ecf9862e6fba509b7e49e3c2511c7d11627/image_geometry/src/pinhole_camera_model.cpp#L299-L308
        fx = np.array(camera_info.p, dtype=float).reshape(3, 4)[0][0]
        fy = np.array(camera_info.p, dtype=float).reshape(3, 4)[1][1]
        cx = np.array(camera_info.p, dtype=float).reshape(3, 4)[0][2]
        cy = np.array(camera_info.p, dtype=float).reshape(3, 4)[1][2]
        tx = np.array(camera_info.p, dtype=float).reshape(3, 4)[0][3]
        ty = np.array(camera_info.p, dtype=float).reshape(3, 4)[1][3]
        ux = (fx * -transformed_point[1] + tx) / transformed_point[0] + cx
        uy = (fy * transformed_point[2] + ty) / transformed_point[0] + cy
        return (ux, uy)


if __name__ == "__main__":
    pass



import numpy as np
import cv2
import sys
from mcap.reader import NonSeekingReader
import amber_mcap.tf2_amber
from amber_mcap.dataset.topic_config import TfTopicConfig
from amber_mcap.dataset.conversion import build_transform_stamped_message
from amber_mcap.exception import TaskDescriptionError, RuntimeError
from typing import List, Tuple, Optional, Dict, cast
from pathlib import Path
import xml.etree.ElementTree as ET
import quaternion


def build_tf_buffer(
    rosbag_files: List[str],
    topic_config: TfTopicConfig = TfTopicConfig(),
    compressed: bool = False,
) -> Tuple[amber_mcap.tf2_amber.BufferCore, Optional[float], Optional[float]]:
    class Origin:
        rpy: Tuple[float, float, float] = (0.0, 0.0, 0.0)
        xyz: Tuple[float, float, float] = (0.0, 0.0, 0.0)

    class FixedJoint:
        origin: Origin = Origin()
        parent: str
        child: str

        def to_tf_message(
            self,
            timestamp: int,
        ) -> amber_mcap.tf2_amber.TransformStamped:
            quat = quaternion.from_euler_angles(
                np.array(
                    [
                        float(self.origin.rpy[0]),
                        float(self.origin.rpy[1]),
                        float(self.origin.rpy[2]),
                    ]
                )
            )
            tf_amber_message = amber_mcap.tf2_amber.TransformStamped(
                amber_mcap.tf2_amber.Header(
                    amber_mcap.tf2_amber.Time(
                        timestamp // 10**9, timestamp % 10**9
                    ),
                    self.parent,
                ),
                self.child,
                amber_mcap.tf2_amber.Transform(
                    amber_mcap.tf2_amber.Vector3(
                        float(self.origin.xyz[0]),
                        float(self.origin.xyz[1]),
                        float(self.origin.xyz[2]),
                    ),
                    amber_mcap.tf2_amber.Quaternion(quat.x, quat.y, quat.z, quat.real),
                ),
            )
            return tf_amber_message

    def load_static_tf_from_urdf_string(
        tf_buffer: amber_mcap.tf2_amber.BufferCore, urdf: str, timestamp: int
    ) -> None:
        def get_fixed_joint(xml: ET.Element) -> FixedJoint:
            joint = FixedJoint()
            for child in xml:
                if child.tag == "origin":
                    if "rpy" in child.attrib:
                        joint.origin.rpy = cast(
                            Tuple[float, float, float],
                            tuple(map(float, child.attrib["rpy"].split(" "))),
                        )
                    if "xyz" in child.attrib:
                        joint.origin.xyz = cast(
                            Tuple[float, float, float],
                            tuple(map(float, child.attrib["xyz"].split(" "))),
                        )
                if child.tag == "parent":
                    joint.parent = child.attrib["link"]
                if child.tag == "child":
                    joint.child = child.attrib["link"]
            return joint

        joints = ET.fromstring(urdf).findall(".//joint")
        for joint in joints:
            if joint.attrib["type"] == "fixed":
                tf_buffer.setTransform(
                    get_fixed_joint(joint).to_tf_message(timestamp),
                    "Authority undetectable",
                    True,
                )

    if topic_config.urdf_path and topic_config.robot_description_topic:
        raise TaskDescriptionError(
            "Do not specidy urdf and robot_description topic at the same time."
        )
    first_timestamp: Optional[int] = None
    last_timestamp: Optional[int] = None
    urdf_string: Optional[str] = None
    tf_buffer = amber_mcap.tf2_amber.BufferCore(
        amber_mcap.tf2_amber.durationFromSec(sys.float_info.max)
    )

    for rosbag_file in rosbag_files:
        reader = NonSeekingReader(rosbag_file)
        for schema, channel, message in reader.iter_messages():
            timestamp = message.log_time

            if first_timestamp is None or timestamp < first_timestamp:
                first_timestamp = timestamp
            if last_timestamp is None or timestamp > last_timestamp:
                last_timestamp = timestamp

            if channel.topic == topic_config.topic_name:
                for tf_amber_message in build_transform_stamped_message(
                    message, schema, compressed
                ):
                    tf_buffer.setTransform(
                        tf_amber_message, "Authority undetectable", False
                    )
            if channel.topic == topic_config.static_tf_topic_name:
                for tf_amber_message in build_transform_stamped_message(
                    message, schema, compressed
                ):
                    tf_buffer.setTransform(
                        tf_amber_message, "Authority undetectable", True
                    )

            if channel.topic == topic_config.robot_description_topic:
                urdf_string = message.data

    if topic_config.urdf_path:
        with open(topic_config.urdf_path, "r", encoding="utf-8") as f:
            urdf_string = f.read()

    if urdf_string:
        if first_timestamp:
            load_static_tf_from_urdf_string(tf_buffer, urdf_string, first_timestamp)
        else:
            raise RuntimeError(
                "Failed to get first timestmp. Please check tf data in rosbag data is exist or not."
            )

    return (tf_buffer, first_timestamp, last_timestamp)
